# DragonBoat结构分析与eBPF结合(Cilium学习)

## 结构分析

### Golang 概述

1. Golang是一种静态编译语言，擅长于后端开发、分布式系统和数据库应用，具有简洁高效的特点。
2. 文件类型
   | 类型     | 描述         |
   | -------- | ------------ |
   | _test.go | 测试代码，与.go文件一一对应 |
   | .go      | 源代码       |
   | .pb.go   | Protocol Buffers生成的代码文件 |
   | .mod     | 依赖模块文件         |
   | .sum     | 依赖模块校验文件       |

3. 文件结构
   | 名称     | 描述 |
   | -------- | ---- |
   | cmd      | 可执行文件，包含不同的package   |
   | internal | 项目内部依赖   |

4. 常用指令链接：[Golang常用指令](https://wnote.com/notes/golang/base/go-command.html)

5. 示例分析：[eBPF-Go指南](https://ebpf-go.dev/guides/getting-started/)
   - `gen.go`: 自定义指令，用bpf2go工具将.c文件转换为.go文件，其中，bpf2go 是一个工具，用于简化和自动化生成 eBPF 程序的 Go 语言绑定
   - `go get`: 获取模块并自动更新go.mod
   - `go mod init <module-name>`: 生成go.mod
   - `go mod tidy`: 自动管理依赖为最新

6. 基本语法
   | 关键词  | 描述 |
   | ------- | ---- |
   | package | 以包为单位，main()入口 |
   | fmt | 用于输出等 |
   | defer | 函数退出前保证执行 |
   | :=与= | 初始化与更新赋值 |
   | `go func() {}` | 创建一个goroutine，匿名函数 |
   | `<-stopper` | 阻塞，直到从stopper通道收到信号 |
   | log | 输出，如`log.Print("...")` |
   | var p Person | 变量类型在后，变量名在前 |
   | `func _CounterClose(closers ...io.Closer) error` | 可变参数，1或多个io.Closer |

7. 重要概念
   - ELF文件在载入内核之前先被划分为Spec (specification)
   - Goroutine是Go语言中的轻量级线程，通过channel实现同步

## eBPF结合DragonBoat示例分析

### Cilium结合DragonBoat示例

1. Cilium uretprobe的main.go
   在头部指定勾连目标：
   ```go
   const (
     binPath = "/bin/bash"
     symbol  = "readline"
   )
   ```

2. 加载入内核：
   ```go
   err := loadBpfObjects(&objs, nil)
   ```

3. 打开用户程序的ELF文件：
   ```go
   ex, err := OpenExecutable(binPath)
   ```

4. 连上用户程序：
   ```go
   err = ex.Uretprobe(symbol, objs.UretprobeBashReadline, nil)
   ```

5. 用户侧读取MAP：
   ```go
   rd, err := perf.NewReader(objs.Events, os.Getpagesize())
   ```

6. 总体编译流程：
   - 编写一个.c文件和main.go文件
   - .c文件被编译成bpfel.o
   - 使用bpf2go工具将bpfel.o自动生成bpfel.go
   - bpfel.go和main.go一起将.o文件分解并载入内核，然后再调几个函数钩到对应位置

### Uprobe read_cmdline的Eunomia和Cilium对比

| Cilium | Eunomia |
|--------|---------|
| main.go指定钩子点 | .bpf.c指定钩子点 |
| *ctx的struct及相应的helper有区别 | - |
| loadBpfObjects(&objs, nil) | 自动完成load pin link |
| 输出是用perf在用户侧输出的 | 输出是直接在内核printk |

### 运行uretprobe示例

要运行这个uretprobe的demo需要amd64架构，可以在服务器上运行：
```sh
wget https://golang.org/dl/go1.22.1.linux-amd64.tar.gz
sudo tar -zxvf go1.22.1.linux-amd64.tar.gz -C /usr/local
export PATH=/usr/local/go/bin:$PATH  # 覆盖旧版本go 1.18.1
cd ebpf/examples/
go run -exec sudo ./uretprobe
```
![跑出demo](https://gitlab.eduxiji.net/T202410558992662/project2210132-233629/-/raw/main/images/cilium_readline_user.png?inline=false)

服务器装了docker可以运行`sudo make`

### 存在的问题与可能的答案

1. **这些example是怎么跑起来的，依赖了什么？**

单独将/xdp放在一个文件夹中：
```sh
go mod init ebpf-xdp
go mod tidy
go get github.com/cilium/ebpf/cmd/bpf2go
go run -exec sudo ./xdp lo
```
可以跑起来，为什么git clone cilium/ebpf之后不需要前三条指令呢？可能是因为向上溯源找到了.go.mod文件？

2. **对照xdp的内核钩子以及uretprobe的编译指令**：
在uprobe中只看到用`__attribute__((unused))`等来定义MAP，xdp和eunomia的uprobe好像可以定义C语言形式的变量和数组，可能只是选择问题。

bpf2go生成_bpfel.o和_bpfeb.o以及.go文件，这些文件虽然帮忙写了些内容，但仍需要自己写main.go来钩上去。main.go大部分结构一致，具体钩的函数有些区别`AttachXDP`。

- [eBPF-Go概念](https://ebpf-go.dev/concepts/loader/)
- [Cilium eBPF示例](https://github.com/cilium/ebpf/blob/main/examples/xdp/bpf_bpfel.go)

3. **ELF文件在载入内核之前，先被分解为了Spec**：
```go
spec, err := ebpf.LoadCollectionSpec("bpf_prog.o")
coll, err := ebpf.NewCollection(spec)
var objs myObjs
err := spec.LoadAndAssign(&objs, nil)
```
查看uretprobe自动生成的el.go，主要是封装了上述几个函数。

### 实践修改uretprobe，钩helloworld

uretprobe的common.h文件在哪？在服务器上单独拿出一个/uretprobe也能运行：
```sh
find ./ -name common.h
```
结果：
```
./ebpf/testdata/common.h
./ebpf/examples/headers/common.h
```
可能go mod tidy功能比较牛，可以自动找到这些文件。

突然间我的`go mod tidy`不工作并显示`i/o timeout`，搜索后我尝试设置了：
```sh
export GOPROXY=direct
```
但是没有用，结果后来自己好了，可能是因为某个服务器宕机。

### DragonBoat结合eBPF的进一步思考

1. 如何根据Dragonboat函数名字钩上去？
2. 哪些Dragonboat的函数可以被钩？
3. 结合libbpf-bootstrap进行更深入的探讨：[libbpf-bootstrap指南](https://nakryiko.com/posts/libbpf-bootstrap/#makefile)
4. 其他概念 fd bpffs pin 不同go进程间共享，进程退出，BPF程序还在
5. go有垃圾回收机制，ProgramArray是个特殊的map，可能会带来bug，建议pin一下